import json
from datetime import datetime
from typing import TypeAlias, Literal

import aiosqlite
from pydantic import BaseModel, field_validator

from browsy import _jobs

AsyncConnection: TypeAlias = aiosqlite.Connection

_INIT_SQL = """
CREATE TABLE IF NOT EXISTS jobs (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    name TEXT NOT NULL,
    input TEXT NOT NULL CHECK(json_valid(input)),
    status TEXT NOT NULL,
    created_at DATETIME NOT NULL DEFAULT (strftime('%Y-%m-%d %H:%M:%f', 'now')),
    updated_at DATETIME,
    worker TEXT
);
CREATE INDEX IF NOT EXISTS idx_jobs_status ON jobs(status);

CREATE TABLE IF NOT EXISTS outputs (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    job_id INTEGER NOT NULL,
    output BLOB,
    FOREIGN KEY (job_id) REFERENCES jobs (id)
);
CREATE INDEX IF NOT EXISTS idx_outputs_job_id ON outputs(job_id);
"""


class DBJob(BaseModel):
    """Represents a job record from the database."""

    id: int
    name: str
    input: dict
    status: _jobs.JobStatus
    created_at: datetime
    updated_at: datetime | None
    worker: str | None

    @field_validator("input", mode="before")
    @classmethod
    def json_str_output(cls, v: str | dict) -> dict:
        if isinstance(v, str):
            return json.loads(v)
        return v


class DBOutput(BaseModel):
    id: int
    job_id: int
    output: bytes | None


async def create_connection(db_path: str) -> AsyncConnection:
    conn = await aiosqlite.connect(db_path)
    conn.row_factory = aiosqlite.Row
    return conn


async def init_db(conn: AsyncConnection) -> None:
    await conn.execute("PRAGMA journal_mode = WAL;")
    await conn.commit()

    await conn.executescript(_INIT_SQL)
    await conn.commit()


async def create_job(
    conn: AsyncConnection,
    name: str,
    input_json: str,
) -> DBJob:
    async with conn.execute(
        """
        INSERT INTO jobs (name, input, status)
        VALUES (?, ?, ?)
        RETURNING id, created_at, updated_at, worker
        """,
        (name, input_json, _jobs.JobStatus.PENDING),
    ) as cursor:
        result = await cursor.fetchone()

    await conn.commit()

    return DBJob(
        name=name,
        input=input_json,
        status=_jobs.JobStatus.PENDING,
        **result,
    )


async def get_job_by_id(
    conn: AsyncConnection,
    id_: int,
) -> DBJob | None:
    async with conn.execute(
        """
        SELECT id, name, input, status, created_at, updated_at, worker
        FROM jobs
        WHERE id = ?
        """,
        (id_,),
    ) as cursor:
        result = await cursor.fetchone()

    return DBJob(**result) if result else None


async def get_job_result_by_job_id(
    conn: AsyncConnection,
    job_id: int,
) -> DBOutput | None:
    async with conn.execute(
        """
        SELECT id, job_id, output
        FROM outputs
        WHERE job_id = ?
        """,
        (job_id,),
    ) as cursor:
        result = await cursor.fetchone()

    return DBOutput(**result) if result else None


async def get_next_job(conn: AsyncConnection, worker: str) -> DBJob | None:
    # Acquires a reserved lock, blocking other write transactions
    await conn.execute("BEGIN IMMEDIATE")

    async with conn.execute(
        f"""
        SELECT id, name, input, status, created_at, updated_at, worker
        FROM jobs
        WHERE status = '{_jobs.JobStatus.PENDING.value}'
        ORDER BY created_at ASC
        LIMIT 1
        """
    ) as cursor:
        result = await cursor.fetchone()

    if not result:
        # Releases the lock
        await conn.rollback()
        return None

    db_job = DBJob(**result)

    await conn.execute(
        f"""
        UPDATE jobs
        SET status = '{_jobs.JobStatus.IN_PROGRESS.value}', updated_at = DATETIME('now'), worker = ?
        WHERE id = ?
        """,
        (worker, db_job.id),
    )
    await conn.commit()
    db_job.status = _jobs.JobStatus.IN_PROGRESS
    db_job.worker = worker

    return db_job


async def update_job_status(
    conn: AsyncConnection,
    job_id: int,
    status: Literal[_jobs.JobStatus.DONE, _jobs.JobStatus.FAILED],
    output: bytes | None,
) -> None:
    await conn.execute(
        """
        UPDATE jobs
        SET status = ?, updated_at = DATETIME('now')
        WHERE id = ?
        """,
        (status, job_id),
    )

    if output:
        await conn.execute(
            """
            INSERT INTO outputs (job_id, output)
            VALUES (?, ?)
            """,
            (job_id, output),
        )

    await conn.commit()



from browsy import BaseJob, Page


class ScreenshotJob(BaseJob):
    NAME = "screenshot"

    url: str | None = None
    html: str | None = None
    full_page: bool = False

    async def execute(self, page: Page) -> bytes:
        if self.url:
            await page.goto(self.url)
        elif self.html:
            await page.set_content(self.html)

        return await page.screenshot(full_page=self.full_page)



import logging
from abc import ABC, abstractmethod
from enum import Enum
from typing import ClassVar
from pathlib import Path

from playwright.async_api import Page
from pydantic import BaseModel, ConfigDict

logger = logging.getLogger(__name__)


class JobStatus(str, Enum):
    PENDING = "pending"
    DONE = "done"
    IN_PROGRESS = "in_progress"
    FAILED = "failed"


def collect_jobs_defs(path: str | Path) -> dict:
    """Collect job class definitions from Python files in the specified path.

    Args:
        path: Path to a Python file or directory containing Python files.
            If a directory is provided, all .py files will be searched recursively.

    Returns:
        A dictionary mapping job names to job classes, where each job class is a subclass
        of BaseJob with a defined NAME class variable.

    Raises:
        ValueError: If the path doesn't exist, no job classes are found, or if there are
            duplicate job names. Also raised if a job class is missing the required NAME
            class variable.
    """
    import importlib.util
    import inspect

    jobs = {}
    path = Path(path) if isinstance(path, str) else path

    if not path.exists():
        raise ValueError(f"Path {path} does not exist")

    logger.info("Selected jobs path: %s", str(path))

    def process_file(file_path: Path) -> None:
        if not file_path.suffix == ".py":
            return

        spec = importlib.util.spec_from_file_location(file_path.stem, file_path)
        if not spec or not spec.loader:
            return

        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

        for _, obj in inspect.getmembers(module, inspect.isclass):
            if issubclass(obj, BaseJob) and obj != BaseJob:
                if not hasattr(obj, "NAME"):
                    raise ValueError(
                        f"Job class {obj.__name__!r} must define a NAME class variable"
                    )
                if obj.NAME in jobs:
                    raise ValueError(
                        f"Duplicated job name {obj.NAME!r}. "
                        "Please ensure all job names are unique."
                    )
                jobs[obj.NAME] = obj

    if path.is_file():
        process_file(path)
    else:
        for file_path in path.rglob("*.py"):
            process_file(file_path)

    if len(jobs) == 0:
        raise ValueError("No job classes found in the specified path")

    logger.info(
        "Found %d job(s) definitions: %s",
        len(jobs),
        ", ".join(sorted(jobs.keys())),
    )

    return jobs


class BaseJob(ABC, BaseModel):
    model_config = ConfigDict(
        frozen=True,
        exclude=["execute"],
        extra="forbid",
    )

    # A unique string identifier for the job type that must be defined by subclasses.
    # This is used to identify and route jobs to the appropriate handler class.
    # For example: NAME = "screenshot" or NAME = "pdf_export"
    NAME: ClassVar[str]

    @abstractmethod
    async def execute(self, page: Page) -> bytes:
        """Execute the job using the provided Playwright page.

        This is the main method that defines the job's behavior. It receives a Playwright's `Page`
        object that can be used to interact with a web browser - navigating to URLs, taking
        screenshots, extracting data, etc.

        This method must be implemented by all job subclasses to define their specific
        automation logic.

        Args:
            page: A Playwright `Page` object representing an automated browser

        Returns:
            bytes: The job's output data in binary format
        """

    async def validate_logic(self) -> bool:
        """Validate job parameters before queueing.

        Since BaseJob inherits from pydantic.BaseModel, basic parameter validation like
        type checking and required fields is handled automatically by Pydantic. This method
        provides additional logical validation that cannot be expressed through Pydantic's
        type system alone.

        For example, if a job accepts either a 'url' or 'html' parameter but not both
        simultaneously, that validation logic should be implemented here since it involves
        checking relationships between multiple fields.

        This method is called before a job is added to the queue. The job will only be
        queued if both Pydantic validation and this logical validation pass.

        Returns:
            bool: True if logical validation passes, False otherwise
        """
        return True



import asyncio
import logging
import os
import string
import random

import click

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)
logger = logging.getLogger("cli")


@click.group()
def cli():
    pass


@cli.command(context_settings={"ignore_unknown_options": True})
@click.argument("uvicorn_args", nargs=-1, type=click.UNPROCESSED)
def server(uvicorn_args: tuple[str]):
    _validate_env_vars()

    import uvicorn

    uvicorn.main(["browsy._server:app"] + list(uvicorn_args))


@cli.command()
@click.option(
    "--name", default=None, help="Worker name (random if not provided)"
)
def worker(name: str | None):
    _validate_env_vars()

    from browsy._worker import start_worker

    worker_name = name or f"worker_{_get_random_chars(8)}"
    try:
        asyncio.run(
            start_worker(
                name=worker_name,
                db_path=os.environ["BROWSY_DB_PATH"],
                jobs_path=["BROWSY_JOBS_PATH"],
            )
        )
    except KeyboardInterrupt:
        logger.info(f"Worker {worker_name} shutting down")


def _get_random_chars(length: int) -> str:
    chars = string.ascii_letters + string.digits
    return "".join(random.choice(chars) for _ in range(length))


def _validate_env_vars() -> None:
    jobs_path = os.environ.get("BROWSY_JOBS_PATH")
    db_path = os.environ.get("BROWSY_DB_PATH")

    if not jobs_path and not db_path:
        raise ValueError(
            "BROWSY_JOBS_PATH and BROWSY_DB_PATH"
            " environment variables are required"
        )


if __name__ == "__main__":
    cli()



import os
from contextlib import asynccontextmanager
from typing import Annotated
from base64 import b64encode
from pathlib import Path

from fastapi import FastAPI, Depends, HTTPException, Request
from pydantic import BaseModel, field_validator

from browsy import _database, _jobs


@asynccontextmanager
async def lifespan(app: FastAPI):
    db_path = os.environ.get("BROWSY_DB_PATH")
    if not db_path:
        raise ValueError("BROWSY_DB_PATH not set")

    jobs_path = os.environ.get("BROWSY_JOBS_PATH", str(Path().absolute()))

    app.state.DB_PATH = db_path
    app.state.JOBS_DEFS = _jobs.collect_jobs_defs(jobs_path)

    conn = await _database.create_connection(db_path)

    try:
        await _database.init_db(conn)
    finally:
        await conn.close()

    yield


app = FastAPI(lifespan=lifespan)


async def get_db(request: Request):
    conn = await _database.create_connection(request.app.state.DB_PATH)

    try:
        yield conn
    finally:
        await conn.close()


class JobRequest(BaseModel):
    name: str
    input: dict


class JobOutput(_database.DBOutput):
    output: str

    @field_validator("output", mode="before")
    @classmethod
    def b64encode_output(cls, value: bytes | None) -> str:
        if value is None:
            return ""

        return b64encode(value).decode()


@app.post("/jobs", response_model=_database.DBJob)
async def create_job(
    request: Request,
    r: JobRequest,
    db_conn: Annotated[_database.AsyncConnection, Depends(get_db)],
):
    jobs_defs: dict[str, type[_jobs.BaseJob]] = request.app.state.JOBS_DEFS
    if r.name not in jobs_defs:
        raise HTTPException(400, "Job with that name is not defined.")

    job = jobs_defs[r.name].model_validate(r.input)
    is_valid = await job.validate_logic()
    if not is_valid:
        raise HTTPException(400, "Job validation failed")

    return await _database.create_job(db_conn, r.name, job.model_dump_json())


@app.get("/jobs/{job_id}", response_model=_database.DBJob)
async def get_job_by_id(
    job_id: int, db_conn: Annotated[_database.AsyncConnection, Depends(get_db)]
):
    job = await _database.get_job_by_id(db_conn, job_id)
    if not job:
        raise HTTPException(404, "Job not found")

    return job


@app.get("/jobs/{job_id}/result", response_model=JobOutput)
async def get_job_result_by_job_id(
    job_id: int, db_conn: Annotated[_database.AsyncConnection, Depends(get_db)]
):
    job = await _database.get_job_by_id(db_conn, job_id)
    if not job:
        raise HTTPException(404, "Job not found")

    if job.status == _jobs.JobStatus.PENDING:
        raise HTTPException(404, "Job is pending")
    if job.status == _jobs.JobStatus.IN_PROGRESS:
        raise HTTPException(404, "Job is in progress")
    if job.status == _jobs.JobStatus.FAILED:
        raise HTTPException(500, "Job failed")

    job_result = await _database.get_job_result_by_job_id(db_conn, job_id)
    if job_result is None:
        raise HTTPException(500, "Job finished, but there's no result")

    return job_result



import asyncio
import logging
import time
import os
from pathlib import Path

from playwright.async_api import (
    async_playwright,
    PlaywrightContextManager,
    Browser,
)
import playwright._impl._errors
import playwright.async_api

from browsy import _database, _jobs

_JOB_POLL_INTERVAL = 5
_HEARTBEAT_LOG_INTERVAL = 600

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)
logger = logging.getLogger("worker-main")


async def worker_loop(
    pw_ctx: PlaywrightContextManager,
    db: _database.AsyncConnection,
    name: str,
    jobs_defs: dict[str, type[_jobs.BaseJob]],
) -> None:
    log = logging.getLogger(name)
    log.info("Ready to work")
    shutdown = False
    last_heartbeat = time.monotonic()
    browser: Browser | None = None
    current_job_task: asyncio.Task | None = None

    try:
        browser = await pw_ctx.chromium.launch(headless=True)

        while not shutdown:
            timeref = time.monotonic()

            try:
                job = await _database.get_next_job(db, worker=name)
                if not job:
                    if timeref - last_heartbeat >= _HEARTBEAT_LOG_INTERVAL:
                        log.info("Worker is alive and polling for jobs")
                        last_heartbeat = timeref

                    await asyncio.sleep(_JOB_POLL_INTERVAL)
                    continue

                last_heartbeat = timeref
                log.info(f"Starting job {job.id} ({job.name!r})")

                try:
                    async with await browser.new_context() as ctx:
                        async with await ctx.new_page() as page:
                            current_job_task = asyncio.create_task(
                                jobs_defs[job.name](**job.input).execute(page)
                            )
                            output = await current_job_task

                except (asyncio.CancelledError, playwright.async_api.Error):
                    log.error(f"Job interrupted, marking {job.id} as failed")
                    await _cancel_task(current_job_task)
                    job.status = _jobs.JobStatus.FAILED
                    await _database.update_job_status(
                        db, job.id, job.status, None
                    )
                    shutdown = True
                    break

                except Exception:
                    log.exception("Job execution failed")
                    job.status = _jobs.JobStatus.FAILED

                else:
                    log.info(f"Job {job.id} is done")
                    job.status = _jobs.JobStatus.DONE

                output = output if job.status == _jobs.JobStatus.DONE else None
                await _database.update_job_status(
                    db, job.id, job.status, output
                )

            except asyncio.CancelledError:
                log.info("Shutting down worker (no jobs interrupted)")
                shutdown = True
                break

    finally:
        await _cancel_task(current_job_task, log=log)

        if browser:
            await _shutdown_browser(browser, log=log)


async def _cancel_task(
    task: asyncio.Task | None, log: logging.Logger | None = None
) -> None:
    log = log or logger
    if task and not task.done():
        task.cancel()
        try:
            await task
        except Exception as e:
            log.debug(f"Canceled task clean up error: {e!r}")


async def _shutdown_browser(
    browser: Browser, timeout: float = 5.0, log: logging.Logger | None = None
) -> None:
    log = log or logger
    try:
        await asyncio.wait_for(browser.close(), timeout=timeout)
    except (asyncio.TimeoutError, playwright.async_api.Error) as e:
        log.warning(f"Failed to close browser gracefully: {e!r}")


async def start_worker(
    name: str, db_path: str, jobs_path: str | None = None
) -> None:
    if not os.path.exists(db_path):
        raise FileNotFoundError(db_path)

    jobs_path = jobs_path or Path().absolute()
    jobs_defs = _jobs.collect_jobs_defs(jobs_path)
    conn = await _database.create_connection(db_path)

    try:
        async with async_playwright() as p:
            await worker_loop(pw_ctx=p, db=conn, name=name, jobs_defs=jobs_defs)
    finally:
        await conn.close()



from typing import Literal

from browsy import BaseJob, Page


class PDFJob(BaseJob):
    NAME = "pdf"

    url: str | None = None
    html: str | None = None
    emulate_media: Literal["null", "print", "screen"] | None = None

    async def execute(self, page: Page) -> bytes:
        if self.url:
            await page.goto(self.url)
        elif self.html:
            await page.set_content(self.html)

        if self.emulate_media:
            await page.emulate_media(self.emulate_media)

        return await page.pdf()



import json
import sys
from base64 import b64decode

import requests


def main() -> int:
    if len(sys.argv) != 3:
        print("Usage: python download.py [job ID] [destination filepath]")
        return 1

    _, job_id, filepath = sys.argv

    resp = requests.get(f"http://localhost:8000/jobs/{job_id}/result")
    if not resp.ok:
        try:
            err = resp.json()
        except json.JSONDecodeError:
            resp.raise_for_status()
        else:
            print(f"Status code {resp.status_code}. Error: {err}")
            return 1

    with open(filepath, "wb") as f:
        f.write(b64decode(resp.json()["output"]))

    print("File saved.")


if __name__ == "__main__":
    raise SystemExit(main())
